@@ -4,7 +4,7 @@ from datetime import datetime
 import singer
 from singer import Transformer, metadata, metrics, utils
 from singer.utils import strftime, strptime_to_utc
-from tap_mambu.transform import transform_json
+from tap_mambu.transform import transform_json, transform_activities
 
 LOGGER = singer.get_logger()
 LOOKBACK_DEFAULT = 14
@@ -134,7 +134,8 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
                   body=None,
                   id_fields=None,
                   parent=None,
-                  parent_id=None):
+                  parent_id=None,
+                  apikey_type=None):
 
 
     # Get the latest bookmark for the stream and set the last_integer/datetime
@@ -160,6 +161,11 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
     total_records = 0 # Initialize total
     record_count = limit # Initialize, reset for each API call
 
+    # Initialize next_max_date and number_last_occurrence parameters used in the request for audit_trail
+    if stream_name == 'audit_trail':
+        next_max_date = static_params['occurred_at[lte]']
+        number_last_occurrence = 0
+
     while record_count == limit: # break out of loop when record_count < limit (or not data returned)
         params = {
             'offset': offset,
@@ -167,6 +173,13 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
             **static_params # adds in endpoint specific, sort, filter params
         }
 
+        if stream_name == 'audit_trail':
+            del params['offset']
+            del params['limit']
+            params['from'] = number_last_occurrence
+            params['size'] = limit
+            params['occurred_at[lte]'] = next_max_date
+
         if bookmark_query_field:
             if bookmark_type == 'datetime':
                 params[bookmark_query_field] = last_datetime
@@ -189,6 +202,7 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
             method=api_method,
             path=path,
             version=api_version,
+            apikey_type=apikey_type,
             params=querystring,
             endpoint=stream_name,
             json=body)
@@ -200,6 +214,9 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
             LOGGER.warning('Stream: {} - NO DATA RESULTS')
             break # NO DATA
 
+        if stream_name == 'audit_trail':
+            data = data['events']
+
         # Transform data with transform_json from transform.py
         #  This function converts camelCase to snake_case for fieldname keys.
         # The data_key may identify array/list of records below the <root> element
@@ -214,6 +231,10 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
             transformed_data = transform_json(data, stream_name)
         elif data_key in data:
             transformed_data = transform_json(data, data_key)[data_key]
+
+        if stream_name == 'activities':
+            transformed_data = transform_activities(transformed_data)
+
         # LOGGER.info('transformed_data = {}'.format(transformed_data))  # TESTING, comment out
         if not transformed_data or transformed_data is None:
             record_count = 0
@@ -287,18 +308,20 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
                             parent_id,
                             child_total_records))
 
-        # Update the state with the max_bookmark_value for the stream
-        if bookmark_field:
-            write_bookmark(state,
-                           stream_name,
-                           sub_type,
-                           max_bookmark_value)
 
         # to_rec: to record; ending record for the batch
         to_rec = offset + limit
         if record_count < limit:
             to_rec = total_records
 
+        if stream_name == 'audit_trail':
+            next_max_date = transformed_data[-1]['occurred_at']
+            index = -2
+            number_last_occurrence = 1
+            while index >= -len(transformed_data) and transformed_data[index]['occurred_at'] == next_max_date:
+                number_last_occurrence += 1
+                index -= 1
+
         LOGGER.info('{} - Synced records: {} to {}'.format(
             stream_name,
             offset,
@@ -308,6 +331,13 @@ def sync_endpoint(client, #pylint: disable=too-many-branches
 
         # End: while record_count == limit
 
+    # Update the state with the max_bookmark_value for the stream
+    if bookmark_field:
+        write_bookmark(state,
+                        stream_name,
+                        sub_type,
+                        max_bookmark_value)
+
     # Return total_records across all batches
     return total_records
 
@@ -366,6 +396,19 @@ def sync(client, config, catalog, state):
     loan_transactions_dttm_str = get_bookmark(state, 'loan_transactions', 'self', start_date)
     loan_transactions_dt_str = transform_datetime(loan_transactions_dttm_str)[:10]
     loan_transactions_dttm = strptime_to_utc(loan_transactions_dt_str)
+
+    clients_dttm_str = get_bookmark(state, 'clients', 'self', start_date)
+    clients_dt_str = transform_datetime(clients_dttm_str)[:10]
+
+    groups_dttm_str = get_bookmark(state, 'groups', 'self', start_date)
+    groups_dt_str = transform_datetime(groups_dttm_str)[:10]
+
+    loan_accounts_dttm_str = get_bookmark(state, 'loan_accounts', 'self', start_date)
+    loan_accounts_dt_str = transform_datetime(loan_accounts_dttm_str)[:10]
+
+    deposit_accounts_dttm_str = get_bookmark(state, 'deposit_accounts', 'self', start_date)
+    deposit_accounts_dt_str = transform_datetime(deposit_accounts_dttm_str)[:10]
+
     lookback_days = int(config.get('lookback_window', LOOKBACK_DEFAULT))
     lookback_date = utils.now() - timedelta(lookback_days)
     if loan_transactions_dttm > lookback_date:
@@ -440,13 +483,25 @@ def sync(client, config, catalog, state):
             'id_fields': ['id']
         },
         'clients': {
-            'path': 'clients',
+            'path': 'clients:search',
             'api_version': 'v2',
-            'api_method': 'GET',
+            'api_method': 'POST',
             'params': {
-                'sortBy': 'lastModifiedDate:ASC',
                 'detailsLevel': 'FULL'
             },
+            'body': {
+                "sortingCriteria": {
+                    "field": "lastModifiedDate",
+                    "order": "ASC"
+                },
+                "filterCriteria": [
+                    {
+                        "field": "lastModifiedDate",
+                        "operator": "AFTER",
+                        "value": clients_dt_str
+                    }
+                ]
+            },
             'bookmark_field': 'last_modified_date',
             'bookmark_type': 'datetime',
             'id_fields': ['id']
@@ -472,13 +527,25 @@ def sync(client, config, catalog, state):
             'id_fields': ['id']
         },
         'deposit_accounts': {
-            'path': 'deposits',
+            'path': 'deposits:search',
             'api_version': 'v2',
-            'api_method': 'GET',
+            'api_method': 'POST',
             'params': {
-                'sortBy': 'lastModifiedDate:ASC',
                 'detailsLevel': 'FULL'
             },
+            'body': {
+                "sortingCriteria": {
+                    "field": "lastModifiedDate",
+                    "order": "ASC"
+                },
+                "filterCriteria": [
+                    {
+                        "field": "lastModifiedDate",
+                        "operator": "AFTER",
+                        "value": deposit_accounts_dt_str
+                    }
+                ]
+            },
             'bookmark_field': 'last_modified_date',
             'bookmark_type': 'datetime',
             'id_fields': ['id'],
@@ -532,26 +599,50 @@ def sync(client, config, catalog, state):
             'id_fields': ['encoded_key']
         },
         'groups': {
-            'path': 'groups',
+            'path': 'groups:search',
             'api_version': 'v2',
-            'api_method': 'GET',
+            'api_method': 'POST',
             'params': {
-                'sortBy': 'lastModifiedDate:ASC',
                 'detailsLevel': 'FULL'
             },
+            'body': {
+                "sortingCriteria": {
+                    "field": "lastModifiedDate",
+                    "order": "ASC"
+                },
+                "filterCriteria": [
+                    {
+                        "field": "lastModifiedDate",
+                        "operator": "AFTER",
+                        "value": groups_dt_str
+                    }
+                ]
+            },
             'bookmark_field': 'last_modified_date',
             'bookmark_type': 'datetime',
             'id_fields': ['id']
         },
         'loan_accounts': {
-            'path': 'loans',
+            'path': 'loans:search',
             'api_version': 'v2',
-            'api_method': 'GET',
+            'api_method': 'POST',
             'params': {
-                'sortBy': 'lastModifiedDate:ASC',
                 'detailsLevel': 'FULL',
                 'paginationDetails': 'ON'
             },
+            'body': {
+                "sortingCriteria": {
+                    "field": "lastModifiedDate",
+                    "order": "ASC"
+                },
+                "filterCriteria": [
+                    {
+                        "field": "lastModifiedDate",
+                        "operator": "AFTER",
+                        "value": loan_accounts_dt_str
+                    }
+                ]
+            },
             'bookmark_field': 'last_modified_date',
             'bookmark_type': 'datetime',
             'id_fields': ['id'],
@@ -643,16 +734,67 @@ def sync(client, config, catalog, state):
             'sub_types': ['ASSET', 'LIABILITY', 'EQUITY', 'INCOME', 'EXPENSE']
         },
         'gl_journal_entries': {
-            'path': 'gljournalentries',
+            'path': 'gljournalentries/search',
             'api_version': 'v1',
-            'api_method': 'GET',
-            'params' : {
-                'from': '{gl_journal_entries_from_dt_str}',
-                'to': '{now_date_str}'
+            'api_method': 'POST',
+            'body': {
+                "filterConstraints": [
+                    {
+                        "filterSelection": "CREATION_DATE",
+                        "filterElement": "BETWEEN",
+                        "value": '{gl_journal_entries_from_dt_str}',
+                        "secondValue": "{now_date_str}"
+                    }
+                ]
             },
             'id_fields': ['entry_id'],
             'bookmark_field': 'booking_date',
             'bookmark_type': 'datetime'
+        },
+        'activities': {
+            'path': 'activities',
+            'api_version': 'v1',
+            'api_method': 'GET',
+            'params' : {
+                'from': '{activities_from_dt_str}',
+                'to': '{now_date_str}'
+            },
+            'id_fields': ['encoded_key'],
+            'bookmark_field': 'timestamp',
+            'bookmark_type': 'datetime'
+        },
+        'index_rate_sources': {
+            'path': 'indexratesources',
+            'api_version': 'v2',
+            'api_method': 'GET',
+            'id_fields': ['encoded_key'],
+            'params': {}
+        },
+        'installments': {
+            'path': 'installments',
+            'api_version': 'v2',
+            'api_method': 'GET',
+            'id_fields': ['encoded_key'],
+            'params': {
+                'dueFrom': '{installments_from_dt_str}',
+                'dueTo': '{now_date_str}'
+            },
+            'bookmark_field': 'last_paid_date',
+            'bookmark_type': 'datetime'
+        },
+        'audit_trail': {
+            'path': 'v1/events',
+            'api_version': 'v1',
+            'api_method': 'GET',
+            'id_fields': [],
+            'apikey_type': 'audit',
+            'params': {
+                'sort_order': 'desc',
+                'occurred_at[gte]': '{audit_trail_from_dt_str}',
+                'occurred_at[lte]': '{audit_trail_to_dt_str}'
+            },
+            'bookmark_field': 'occurred_at',
+            'bookmark_type': 'datetime'
         }
     }
 
@@ -666,9 +808,17 @@ def sync(client, config, catalog, state):
     last_stream = singer.get_currently_syncing(state)
     LOGGER.info('last/currently syncing stream: {}'.format(last_stream))
 
+    # Start syncing from last/currently syncing stream
+    if last_stream in selected_streams:
+        selected_streams = selected_streams[selected_streams.index(last_stream):] + selected_streams[:selected_streams.index(last_stream)]
+
     # For each endpoint (above), determine if the stream should be streamed
     #   (based on the catalog and last_stream), then sync those streams.
-    for stream_name, endpoint_config in endpoints.items():
+    for stream_name in selected_streams:
+        endpoint_config = endpoints.get(stream_name)
+        if endpoint_config is None:
+            # null endpoint_config signifies that this is a child stream
+            continue
         should_stream, last_stream = should_sync_stream(selected_streams,
                                                         last_stream,
                                                         stream_name)
@@ -682,8 +832,61 @@ def sync(client, config, catalog, state):
                 # Now date
                 if stream_name == 'gl_journal_entries':
                     now_date_str = strftime(utils.now())[:10]
-                    gl_journal_entries_from_dttm_str = get_bookmark(state, 'gl_journal_entries', sub_type, start_date)
-                    gl_journal_entries_from_dt_str = transform_datetime(gl_journal_entries_from_dttm_str)[:10]
+                    gl_journal_entries_from_dttm_str = get_bookmark(
+                        state, 'gl_journal_entries', sub_type, start_date)
+                    gl_journal_entries_from_dt_str = transform_datetime(
+                        gl_journal_entries_from_dttm_str)[:10]
+                    gl_journal_entries_from_param = endpoint_config.get(
+                        'body', {}).get('filterConstraints', {})[0].get('value')
+                    if gl_journal_entries_from_param:
+                        endpoint_config['body']['filterConstraints'][0]['value'] = gl_journal_entries_from_dt_str
+                    gl_journal_entries_to_param = endpoint_config.get(
+                        'body', {}).get('filterConstraints', {})[0].get('secondValue')
+                    if gl_journal_entries_to_param:
+                        endpoint_config['body']['filterConstraints'][0]['secondValue'] = now_date_str
+
+                if stream_name == 'activities':
+                    now_date_str = strftime(utils.now())[:10]
+                    activities_from_dttm_str = get_bookmark(
+                        state, 'activities', sub_type, start_date)
+                    activities_from_dt_str = transform_datetime(
+                        activities_from_dttm_str)[:10]
+                    activities_from_param = endpoint_config.get(
+                        'params', {}).get('from')
+                    if activities_from_param:
+                        endpoint_config['params']['from'] = activities_from_dt_str
+                    activities_to_param = endpoint_config.get(
+                        'params', {}).get('to')
+                    if activities_to_param:
+                        endpoint_config['params']['to'] = now_date_str
+
+                if stream_name == 'installments':
+                    now_date_str = strftime(utils.now())[:10]
+                    installments_from_dttm_str = start_date
+                    installments_from_dt_str = transform_datetime(
+                        installments_from_dttm_str)[:10]
+                    installments_from_param = endpoint_config.get(
+                        'params', {}).get('dueFrom')
+                    if installments_from_param:
+                        endpoint_config['params']['dueFrom'] = installments_from_dt_str
+                    installments_to_param = endpoint_config.get(
+                        'params', {}).get('dueTo')
+                    if installments_to_param:
+                        endpoint_config['params']['dueTo'] = now_date_str
+
+                if stream_name == 'audit_trail':
+                    now_date_str = strftime(utils.now())
+                    audit_trail_from_dttm_str = get_bookmark(
+                        state, 'audit_trail', sub_type, start_date)
+                    audit_trail_from_dt_str = transform_datetime(
+                        audit_trail_from_dttm_str)
+                    audit_trail_from_param = endpoint_config.get(
+                        'params', {}).get('occurred_at[gte]')
+                    if audit_trail_from_param:
+                        endpoint_config['params']['occurred_at[gte]'] = audit_trail_from_dt_str
+                    audit_trail_to_dt_str = endpoint_config.get('params', {}).get('occurred_at[lte]')
+                    if audit_trail_to_dt_str:
+                        endpoint_config['params']['occurred_at[lte]'] = now_date_str
 
                 update_currently_syncing(state, stream_name)
                 path = endpoint_config.get('path')
@@ -691,13 +894,6 @@ def sync(client, config, catalog, state):
                 if sub_type_param:
                     endpoint_config['params']['type'] = sub_type
 
-                gl_journal_entries_from_param = endpoint_config.get('params', {}).get('from')
-                if gl_journal_entries_from_param:
-                    endpoint_config['params']['from'] = gl_journal_entries_from_dt_str
-                gl_journal_entries_to_param = endpoint_config.get('params', {}).get('to')
-                if gl_journal_entries_to_param:
-                    endpoint_config['params']['to'] = now_date_str
-
                 total_records = sync_endpoint(
                     client=client,
                     catalog=catalog,
@@ -715,7 +911,9 @@ def sync(client, config, catalog, state):
                     bookmark_type=endpoint_config.get('bookmark_type'),
                     data_key=endpoint_config.get('data_key', None),
                     body=endpoint_config.get('body', None),
-                    id_fields=endpoint_config.get('id_fields'))
+                    id_fields=endpoint_config.get('id_fields'),
+                    apikey_type=endpoint_config.get('apikey_type', None)
+                )
 
                 update_currently_syncing(state, None)
                 LOGGER.info('Synced: {}, total_records: {}'.format(
